#include <iostream>
#include <getopt.h>
#include <cstdlib>
#include <iterator>
#include <fstream>
#include "config.h"
#include "bitops.hpp"
#include "Hypergraph.hpp"
#include "db.hpp"

using namespace std;

void usage(const char *th) {
  cout << hslinkern_NAME <<", (C) 2011 Rene van Bevern <rene.vanbevern@tu-berlin.de>" << endl
       << "-------------------------------------------------------------------------" << endl;
  cout << "Usage: " << th << " [options] hypergraph_file.txt" << endl;
  cout << endl << "Valid options are" << endl;
  cout << "-k K: kernelize the instance asking for a hitting set of size K." << endl;
  cout << "      Per default, an upper bound on the minimum hitting set size" << endl;
  cout << "      is used as K." << endl;
  cout << "-o OUTPUT: write kernelized hypergraph into the file OUTPUT." << endl;
  cout << "           By default, no output is generated, only statistics." << endl;
  cout << "-s: find the optimum hitting set in the kernelized instance using" << endl;
  cout << "    an exponential-time exact search-tree algorithm, " << endl;
  cout << "      time O(d^k + |V| + |E|)" << endl;
  exit(1);
}    

struct petals_less_than {
  bool yes;
  unsigned int k;

  petals_less_than(int _k) : k(_k) { yes = true; }
  void operator()(const Edge &c) {
    if(lookup(c)->petals > k) yes = false;
  }
};

struct add_petal {
  const Edge &e;
  add_petal(const Edge &_e) : e(_e) {}
  void operator()(const Edge &c) {
    Leaf *l(lookup(c));
    Edge enotc(edge_subtraction(e, c));
    
    for(Edge::const_iterator i = enotc.begin(); i != enotc.end(); ++i)
      if(l->used(*i)) return; // l->forbidden = true;

    //    if(l->forbidden) return; // XXX what is this for?

    l->petals++;

    // cout << "petals " << l->petals << " for ";
    // copy(c.begin(), c.end(), ostream_iterator<Vertex>(cout, " "));
    // cout << "(added ";
    // copy(e.begin(), e.end(), ostream_iterator<Vertex>(cout, " "));
    // cout << ") and enotc is ";
    // copy(enotc.begin(), enotc.end(), ostream_iterator<Vertex>(cout, " "));
    // cout << endl;

    //    l->forbidden = true;

    for(Edge::const_iterator i = enotc.begin(); i != enotc.end(); ++i)
      l->setused(*i);
  }
};

Hypergraph kernelize(const Hypergraph &G, const Graphstats &s, unsigned int k) {
  Hypergraph H;

  cout << endl << "... computing kernel for k = " << k << " ..." << endl;
  
  for(Hypergraph::const_iterator i = G.begin(); i != G.end(); ++i) {
    petals_less_than lt(k);
    foreach_intersection(G, *i, s, lt);
    if(lt.yes) {
      add_petal addp(*i);
      H.push_back(*i);
      foreach_intersection(G, *i, s, addp);
      //      foreach_intersection(G, *i, s, unforbid);
      // cout << "petals " << k+1 << " for ";
      // copy(i->begin(), i->end(), ostream_iterator<Vertex>(cout, " "));
      // cout << endl;
      lookup(*i)->petals = k+1;
    // } else {
    //   cout << "threw away ";
    //   copy(i->begin(), i->end(), ostream_iterator<Vertex>(cout, " "));
    //   cout << endl;
    }

    if(lookup(Edge())->petals > k) {
      cout << "No-instance" << endl;
      break;
    }
  }
  return H;
}

void print_stats(const Graphstats &stats) {
  cout << "----------------" <<endl;
  cout << "Edges              : " << stats.edges << endl;
  cout << "Maximum edge size  : " << stats.edgesize << endl;
  cout << "Smallest huge edge : " << stats.hugesize << endl;
  cout << "Huge edges         : " << stats.hugeedges << endl;
}

int main(int argc, char *argv[]) {
  int c;
  bool searchtree = false;
  int k = -1;
  ofstream of;
  
  while((c=getopt(argc, argv, "k:o:s")) != -1)
    switch (c) {
    case 'k': {
      k = atoi(optarg);
      assert(k > 0);
      break;
    }
    case 'o': {
      of.open(optarg);
      break;
    }
    case 's': {
      searchtree = true;
      break;
    }
    default: usage(argv[0]);
    }
  
  if(optind >= argc) {
    cout << "Error: no filename given" << endl;
    usage(argv[0]);
  }

  cout << "Reading hypergraph " << argv[optind] << endl;
  Hypergraph G(read_hypergraph(argv[optind]));
  cout << "Graph statistics" <<endl;
  Graphstats stats(get_stats(G));
  print_stats(stats);

  edges_by_size(G, stats);
  Hypergraph apx(approx_hs(G, stats));
  

  int apxbound = get_stats(apx).edgesums;
  int greedybound = greedy_hs(G, stats);
  if(k < 0) {
    k = apxbound;
    if(greedybound < k) k = greedybound;
  }

  cout << "Upper bound HS     : " << k << endl;
  cout << "Lower bound HS     : " << apx.size() << endl;

  initialize_db(G, stats);

    //    return 0;
    //  }
  Hypergraph H(kernelize(G, stats, k));
  Graphstats hstats(get_stats(H));
  cout << endl << "Kernel statistics" << endl;
  print_stats(hstats);

  if(searchtree) {
    vector<bool> hit(hstats.vertices+1);
    cout << endl << "... solving ..." << endl;
    cout << endl << "Optimal solution size: " << searchtree_hs(H,hstats,hit,k) << endl;
  }

  if(of.is_open()) {
    for(Hypergraph::iterator i = H.begin(); i != H.end(); ++i) {
      copy(i->begin(), i->end(), ostream_iterator<Vertex>(of, " "));
      of << endl;
    }
    of.close();
  }
  return 0;
}
